package DumpComposer;

use strict;
use XmlNode;
use Logging;
use Agent;
use MigrationDumpStatus;


#
# Constants
#
my $DUMP_FILE_NAME = 'dump.xml';
my $STATUS_FILE_NAME = 'dumping-status.xml';
my $CONTENT_LIST_FILE_NAME = 'content-list.xml';

#
# make dump in 'shallow' mode. Corresponds to old "-da -nc" mode
#
sub makeShallowDump {
  my $workDir = shift;

  Agent::setWorkDir($workDir);

  my $migrationDumpNode = &createMigrationDumpNode();

  # dump resellers
  my @resellerNodes = map { _makeResellerShallowNode ($_) } Agent::getResellers();
  $migrationDumpNode->addChildren(@resellerNodes);

  # dump admin clients
  my @adminClientNodes = map { _makeClientShallowNode ($_) } Agent::getClients( undef );
  $migrationDumpNode->addChildren(@adminClientNodes);

  # dump admin domains
  my @adminDomainNodes = map { _makeDomainShallowNode ($_) } Agent::getDomains( undef );
  $migrationDumpNode->addChildren(@adminDomainNodes);

  unless ( &serializeXmlNode($migrationDumpNode,"$workDir/$DUMP_FILE_NAME") ) {
    Logging::error("Dump file '$DUMP_FILE_NAME' is not saved.");
  }
  return $migrationDumpNode;
}

sub _makeResellerShallowNode {
  my ( $resellerId ) = @_;

  my $resellerNode = XmlNode->new('reseller', (
    'attributes' => {
      'name' => $resellerId,
      'guid' => ''
    }, 
    'children' => [
      XmlNode->new('preferences'),
      _makeFakeClientPropertiesNode(),
    ]
  ));
      
  my @clientIds = Agent::getClients($resellerId);
  if (@clientIds) {
    my @clientNodes = map { _makeClientShallowNode ($_) } @clientIds;
    $resellerNode->addChild(
      XmlNode->new('clients', ('children' => \@clientNodes))
    );
  }

  my @domainIds = Agent::getDomains($resellerId);
  if (@domainIds) {
    my @domainNodes = map { _makeDomainShallowNode ($_) } @domainIds;
    $resellerNode->addChild(
      XmlNode->new('domains', ('children' => \@domainNodes))
    );
  }

  return $resellerNode;
}

sub _makeClientShallowNode {
  my ( $clientId ) = @_;

  my $clientNode = XmlNode->new('client', (
    'attributes' => {
      'name' => $clientId,
      'guid' => ''
    },
    'children' => [
      XmlNode->new('preferences'),
      _makeFakeClientPropertiesNode(),
    ]
  ));

  my @domainIds = Agent::getDomains($clientId);
  if (@domainIds) {
    my @domainNodes = map { _makeDomainShallowNode ($_ ) } @domainIds;
    $clientNode->addChild(
      XmlNode->new('domains', ('children' => \@domainNodes))
    );
  }

  return $clientNode;
}

sub _makeDomainShallowNode {
  my ( $domainId ) = @_;

  return XmlNode->new('domain', (
    'attributes' => {
      'name' => $domainId,
      'guid' => ''
    },
    'children' => [
      XmlNode->new('preferences'),
      Agent::getDomainProperties($domainId)
    ]
  ));
}

#
# Pass 'undef' for $resellers, $clients and $domains to make 'dump-all'
#
sub makeDump {
  my ( $workDir, $resellers, $clients, $domains, $configOnly, $onlyMail, $onlyHosting, $gzip) = @_;

  Agent::setWorkDir($workDir);

  $resellers = &getAllResellers() unless( defined ( $resellers ) );
  $clients = &getAllClients() unless( defined ( $clients ) );
  $domains = &getAllDomains() unless( defined ( $domains ) );
  
  my $dumpOptions = {
    'configOnly' => $configOnly,
    'onlyMail' => $onlyMail,
    'onlyHosting'=> $onlyHosting,
    'gzip' => $gzip
  };
  
  my $statusFile = "$workDir/$STATUS_FILE_NAME";
  my $dumpStatus = MigrationDumpStatus->new($statusFile);
  $dumpStatus->start(scalar(@{$resellers}) + scalar(@{$clients}), scalar(@{$domains}));

  my $migrationDumpNode = &createMigrationDumpNode($configOnly);

  _addDataToMigrationDumpNode( $migrationDumpNode, $resellers, $clients, $domains, $dumpOptions );

  my $contentListNode = &getContentList($migrationDumpNode);

  if ( &serializeXmlNode($migrationDumpNode,"$workDir/$DUMP_FILE_NAME") ) {
    my $migrationDumpFileSize = -s "$workDir/$DUMP_FILE_NAME";

    if ( $migrationDumpFileSize ) {
      my $fileNode = XmlNode->new( 'file' );
      $fileNode->setAttribute( 'size', $migrationDumpFileSize );
      $fileNode->setText( $DUMP_FILE_NAME );
      $contentListNode->addChild( $fileNode );

      my $contentListSize = $contentListNode->getAttribute( 'size' ) || '0';
      my $contentListNewSize = int($contentListSize) + int($migrationDumpFileSize);
      $contentListNode->setAttribute( 'size', $contentListNewSize );
    }
  }
  else {
    Logging::error("Dump file '$DUMP_FILE_NAME' is not saved.");
  }

  unless ( &serializeXmlNode($contentListNode,"$workDir/$CONTENT_LIST_FILE_NAME") ) {
    Logging::error("Content list file '$CONTENT_LIST_FILE_NAME' is not saved.");
  }

  return $migrationDumpNode;
}

sub _addDataToMigrationDumpNode {
  my ( $migrationDumpNode, $resellers, $clients, $domains, $dumpOptions ) = @_;

  Logging::trace('Dumping resellers...');
  foreach my $resellerId (@{$resellers}) {
    my $resellerNode = Agent::getReseller($resellerId);
    
    Logging::trace("Dumping '$resellerId' reseller's domains...");
    my @resDomainIds = Agent::getDomains($resellerId);
    if (@resDomainIds) {
      my @domainNodes = map { _makeDomainNode ($_, $dumpOptions ) } @resDomainIds; 
      $resellerNode->addChild(
        XmlNode->new('domains', ('children' => \@domainNodes)) 
      ); 
    }

    Logging::trace("Dumping '$resellerId' reseller's clients...");
    my @resClientIds = Agent::getClients($resellerId);
    if (@resClientIds) {
      my @clientNodes = map { _makeClientNode($_, $domains, $dumpOptions ) } @resClientIds;
      $resellerNode->addChild(
        XmlNode->new('clients', ('children' => \@clientNodes))
      );
    }
  
    $migrationDumpNode->addChild($resellerNode);
  }

  Logging::trace("Dumping root clients...");
  my @clientIds = Agent::getClients(undef);
  if (@clientIds) {
    $migrationDumpNode->addChildren(
      map { _makeClientNode($_, $domains, $dumpOptions ) } @clientIds
    );
  }

  Logging::trace("Dumping root domains...");
  my @domainIds = Agent::getDomains(undef);
  if (@domainIds) {
    $migrationDumpNode->addChildren(
      map { _makeDomainNode($_, $dumpOptions) } @domainIds
    );
  }
}

sub getAllResellers {
  my @resellerIds = Agent::getResellers();
  return \@resellerIds;
}

sub getAllClients {
  my @clientIds; 

  my @resellerIds = Agent::getResellers();
  foreach my $resellerId ( @resellerIds ) {
    my @resClientIds = Agent::getClients( $resellerId );
    if ( @resClientIds ) {
      push (@clientIds, @resClientIds); 
    }
  }
  
  my @rootClientIds = Agent::getClients( undef );
  push (@clientIds, @rootClientIds); 
  
  return \@clientIds;
}

sub getAllDomains {
  my @domainIds; 

  my @resellerIds = Agent::getResellers();
  foreach my $resellerId ( @resellerIds ) {
    my @resClientIds = Agent::getClients( $resellerId );
    foreach my $resClientId ( @resClientIds ) {
      my @clDomainIds = Agent::getDomains( $resClientId );
      if ( @clDomainIds ) {
        push (@domainIds, @clDomainIds);
      }
    }
    my @resDomainIds = Agent::getDomains( $resellerId );
    if ( @resDomainIds ) {
      push (@domainIds, @resDomainIds);
    }
  }
  
  my @rootDomains = Agent::getDomains( undef );
  push (@domainIds, @rootDomains); 
  
  return \@domainIds;
}

sub createMigrationDumpNode {
  my $configOnly = shift;
  my $migrationDumpNode = XmlNode->new( 'migration-dump' );
  $migrationDumpNode->setAttribute( 'agent-name', Agent::getAgentName() );
  $migrationDumpNode->setAttribute( 'dump-version', Agent::getAgentVersion() );
  $migrationDumpNode->setAttribute( 'content-included', $configOnly ? 'false' : 'true' );

  my $dumpInfoNode = XmlNode->new( 'dump-info' );

  my $cpDescriptionNode = XmlNode->new( 'cp-description' );
  $cpDescriptionNode->setAttribute( 'type', Agent::getAgentName() );
  $dumpInfoNode->addChild( $cpDescriptionNode );

  my $contentTransportDescriptionNode = Agent::getContentTransportDescription();
  if ( $contentTransportDescriptionNode =~ /XmlNode/ ) {
    my $contentTransportNode = XmlNode->new( 'content-transport' );
    $contentTransportNode->addChild( $contentTransportDescriptionNode );
    $dumpInfoNode->addChild ( $contentTransportNode );
  }

  $migrationDumpNode->addChild( $dumpInfoNode );
  return $migrationDumpNode;
}

sub _makeClientNode {
  my ( $clientId, $domains, $dumpOptions ) = @_;

  my $clientNode = Agent::getClient( $clientId, $dumpOptions );
  Logging::trace("Dumping '$clientId' client's domains...");
  my @clDomainIds = Agent::getDomains($clientId);
  if (@clDomainIds) {
    my @domainNodes = map { _makeDomainNode($_, $dumpOptions) } @clDomainIds;
    $clientNode->addChild(
      XmlNode->new('domains', ('children' => \@domainNodes))
    );
  }
  return $clientNode;
}

sub _makeDomainNode {
  my ($domainId, $dumpOptions) = @_;

  # The dump returned in 'recursive' mode doesn't have content and is always 'full'.
  my $recursive = 'recursive';

  my %contentOptions;
  $contentOptions{'only-mail'} = 1 if defined ( $dumpOptions->{'onlyMail'} );
  $contentOptions{'only-hosting'} = 1 if defined ( $dumpOptions->{'onlyHosting'} );

  my $domainNode = Agent::getDomain( $domainId, $recursive, $dumpOptions->{'configOnly'} ? undef : 'content', $dumpOptions->{'configOnly'} ? undef : $dumpOptions->{'gzip'}, \%contentOptions);

  unless ( ref($domainNode) =~ /XmlNode/ ) {return undef};
  unless ($domainNode->getName() eq 'domain' ) {return undef};

  unless ( defined ( $dumpOptions->{'onlyMail'} ) || defined ( $dumpOptions->{'onlyHosting'} ) ) {
    return $domainNode;
  }

  #
  # Process 'onlyMail' and 'onlyHosting' cases.
  # 
  if ( $dumpOptions->{'onlyMail'} ) {
    my $onlyMailDomain = $domainNode->copy('no_children');

    $onlyMailDomain->addChild( XmlNode->new ( 'preferences' ) );

    my $onlyMailPropertiesNode = XmlNode->new( 'properties' );
    my $propertiesNode = $domainNode->getChild( 'properties' );
    my $ipNode = $propertiesNode->getChild( 'ip' );
    $onlyMailPropertiesNode->addChild( $ipNode ) if defined ( $ipNode );
    $onlyMailPropertiesNode->addChild( $propertiesNode->getChild( 'status' ) );

    $onlyMailDomain->addChild( $onlyMailPropertiesNode );

    my $mailsystemNode = $domainNode->getChild( 'mailsystem' );
    $onlyMailDomain->addChild( $mailsystemNode ) if defined ( $mailsystemNode );

    my $maillistsNode = $domainNode->getChild( 'maillists' );
    $onlyMailDomain->addChild( $maillistsNode ) if defined ( $maillistsNode );

    return $onlyMailDomain;
  }
  elsif ( $dumpOptions->{'onlyHosting'} ) {
    my $onlyHostingDomain = $domainNode->copy('no_children');

    $onlyHostingDomain->addChild( $domainNode->getChild( 'preferences' ) );

    $onlyHostingDomain->addChild( $domainNode->getChild( 'properties' ) );

    my $limitsAndPermissionsNode = $domainNode->getChild( 'limits-and-permissions' );
    $onlyHostingDomain->addChild( $limitsAndPermissionsNode ) if defined ( $limitsAndPermissionsNode );

    my $databasesNode = $domainNode->getChild( 'databases' );
    $onlyHostingDomain->addChild( $databasesNode ) if defined ( $databasesNode );

    my @odbcDsnNodes = $domainNode->getChildren( 'odbc-dsn' );
    if ( @odbcDsnNodes ) {
      foreach my $odbcDsnNode ( @odbcDsnNodes ) {
        $onlyHostingDomain->addChild( $odbcDsnNode );
      }
    }

    my @coldfusionDsnNodes = $domainNode->getChildren( 'coldfusion-dsn' );
    if ( @coldfusionDsnNodes ) {
      foreach my $coldfusionDsnNode ( @coldfusionDsnNodes ) {
        $onlyHostingDomain->addChild( $coldfusionDsnNode );
      }
    }

    my $trafficNode = $domainNode->getChild( 'traffic' );
    $onlyHostingDomain->addChild( $trafficNode ) if defined ( $trafficNode );

    my $certificatesNode = $domainNode->getChild( 'certificates' );
    $onlyHostingDomain->addChild( $certificatesNode ) if defined ( $certificatesNode );

    my $tomcatNode = $domainNode->getChild( 'tomcat' );
    $onlyHostingDomain->addChild( $tomcatNode ) if defined ( $tomcatNode );

    my $domainuserNode = $domainNode->getChild( 'domainuser' );
    $onlyHostingDomain->addChild( $domainuserNode ) if defined ( $domainuserNode );

    my $phostingNode = $domainNode->getChild( 'phosting' );
    if ( defined ( $phostingNode ) ) {
      $onlyHostingDomain->addChild( $phostingNode );
    }
    else {
      my $shostingNode = $domainNode->getChild( 'shosting' );
      if ( defined ( $shostingNode ) ) {
        $onlyHostingDomain->addChild( $shostingNode );
      }
      else {
        my $fhostingNode = $domainNode->getChild( 'fhosting' );
        if ( defined ( $fhostingNode ) ) {
          $onlyHostingDomain->addChild( $fhostingNode );
        }
      }
    }

    return $onlyHostingDomain;
  }

	return $domainNode;
}

sub _makeFakeClientPropertiesNode {
  my $propertiesNode = XmlNode->new('properties');
  my $passwordNode   = XmlNode->new('password');
  $passwordNode->setAttribute( 'type', 'empty' );

  my $statusNode = XmlNode->new('status');
  $statusNode->addChild( XmlNode->new('enabled') );
  $propertiesNode->addChild($passwordNode);
  $propertiesNode->addChild($statusNode);
  return $propertiesNode;
}


#
# Store $xmlNode to $filename
#
sub serializeXmlNode {
  my ($xmlNode, $filename) = @_;
  
  unless ( open DUMP, ">$filename" ) {
    Logging::error("Unable open file '". $filename . "' for write: $!");
    return;
  }
  binmode DUMP;
  $xmlNode->serialize(\*DUMP);
  close DUMP;
  return 1;
}
#
# Returns an XmlNode with content list.
#
sub getContentList {
  my $dumpXmlNode = shift;
  unless ( ref($dumpXmlNode) =~ /XmlNode/ ) {return undef};
  
  my $contentListNode = XmlNode->new( 'contentlist' );
  $contentListNode->setAttribute( 'size', '0');
  
  &_lookup4ContentNode($dumpXmlNode, $contentListNode);
  
  return $contentListNode;
}

sub _lookup4ContentNode {
  my ($node, $contentListNode) = @_;

  my @childNodes = $node->getChildren( undef );
  if ( @childNodes ) {
    foreach my $childNode ( @childNodes ) {
      if ( $childNode->getName() eq 'content' ) {
        &_processContentNode($childNode, $contentListNode);
      }
      else {
        &_lookup4ContentNode($childNode, $contentListNode);
      }
    }
  }
}

sub _processContentNode {
  my ($node, $contentListNode) = @_;

  my @cidNodes = $node->getChildren( 'cid' );
  foreach my $cidNode ( @cidNodes ) {
    my $path = $cidNode->getAttribute( 'path' );
    $path = '' unless ( defined ( $path ) );
    $path .= '/' if $path and substr( $path, -1, 1 ) ne '/';
    my @contentFileNodes = $cidNode->getChildren( 'content-file' );
    foreach my $contentFileNode ( @contentFileNodes ) {
      my $size = $contentFileNode->getAttribute( 'size' );
      my $filename = $contentFileNode->getText();
      
      my $fileNode = XmlNode->new( 'file' );
      $fileNode->setAttribute( 'size', $size || '0');
      $fileNode->setText( $path . $filename );
      
      $contentListNode->addChild( $fileNode );
      
      if ( defined ( $size ) ) {
        my $contentListSize = $contentListNode->getAttribute( 'size' ) || '0';
        my $contentListNewSize = int($contentListSize) + int($size);
        $contentListNode->setAttribute( 'size', $contentListNewSize );
      }
    }
  }
}

1;
